/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Uwe Schulzweida

*/

#include <cdi.h>

#include <cassert>
#include <algorithm>

#include "percentiles.h"
#include "varray.h"
#include "field.h"
#include "cdo_output.h"
#include "field_functions.h"


void
field_copy(const Field_t &field_src, Field_t &field_tgt)
{
  auto func = [](const LambdaField_T_ f_src, LambdaField_T_ f_tgt) {
    f_tgt.nmiss = f_src.nmiss;
    f_tgt.vec = f_src.vec;
  };
  
  field_operation2(func, field_src, field_tgt);
}

// void
// field_copy(const Field3D_t &field_src, Field3D_t &field_tgt)
// {
//   field_tgt.nmiss = field_src.nmiss;

//   if (field_src.memType == MemType::Float)
//     field_tgt.vec_f = field_src.vec_f;
//   else
//     field_tgt.vec_d = field_src.vec_d;
// }

// void
// field_copy(const Field3D_t &field_src, const int levelID, Field_t &field_tgt)
// {
//   const auto size = field_src.gridsize * field_src.nwpv;
//   const auto offset = levelID * size;
//   if (field_src.memType == MemType::Float)
//     std::copy(field_src.vec_f.begin() + offset, field_src.vec_f.begin() + offset + size, field_tgt.vec_f.begin());
//   else
//     std::copy(field_src.vec_d.begin() + offset, field_src.vec_d.begin() + offset + size, field_tgt.vec_d.begin());
// }

// void
// field_add(Field_t &field1, const Field3D_t &field2, const int levelID)
// {
//   const auto size = field1.gridsize * field1.nwpv;
//   const auto offset = levelID * size;
//   if (field1.memType == MemType::Float)
//     for (size_t i = 0; i < size; ++i) field1.vec_f[i] += field2.vec_f[offset + i];
//   else
//     for (size_t i = 0; i < size; ++i) field1.vec_d[i] += field2.vec_d[offset + i];
// }

// functor that returns true if value is equal to the value of the constructor parameter provided
class valueDblIsEqual
{
  double _missval;

public:
  explicit valueDblIsEqual(double missval) : _missval(missval) {}
  bool
  operator()(const double value) const
  {
    return dbl_is_equal(value, _missval);
  }
};

// functor that returns true if value is equal to the value of the constructor parameter provided
class valueIsEqual
{
  double _missval;

public:
  explicit valueIsEqual(double missval) : _missval(missval) {}
  bool
  operator()(const double value) const
  {
    return is_equal(value, _missval);
  }
};

size_t
field_num_miss(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    field.check_gridsize();

    const auto missval = field.missval;
    const auto &v = field.vec;
    return std::count_if(v.begin(), v.begin() + field.size, valueDblIsEqual(missval));
  };
  

  return field_operation(func, field);
}

size_t
field_num_mv(Field_t &field)
{ 
  return field_operation(&field_num_mv<float>, &field_num_mv<double>, field);
}

MinMax
field_min_max(const Field_t &field)
{
    const auto func = [](const LambdaField_T_ field) {
      if(field.nmiss) {
        return varray_min_max_mv(field.size, field.vec, field.getMissValue());
      } else {
        return varray_min_max(field.vec);
      }
    };
  
  return field_operation(func, field);
}

double
field_min(const Field_t &field)
{
    const auto func = [](const LambdaField_T_ field) -> double {
      if(field.nmiss) {
        return varray_min_mv(field.size, field.vec, field.getMissValue());
      } else {
        return varray_min(field.size, field.vec);
      }
    };
  
  return field_operation(func, field);
}

double
field_max(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) -> double {
    if(field.nmiss) {
      return varray_max_mv(field.size, field.vec, field.getMissValue());
    } else {
      return varray_max(field.size, field.vec);
    }
  };
  
  return field_operation(func, field);
}

double
field_range(const Field_t &field)
{
const auto func = [](const LambdaField_T_ field) -> double {
    if(field.nmiss) {
      return varray_range_mv(field.size, field.vec, field.getMissValue());
    } else {
      return varray_range(field.size, field.vec);
    }
  };
  
  return field_operation(func, field);
}

double
field_sum(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    if(field.nmiss) {
      return varray_sum_mv(field.size, field.vec, field.getMissValue());
    } else {
      return varray_sum(field.size, field.vec);
    }
  };
  
  return field_operation(func, field);
}

double
field_mean(const Field_t &field)
{
   const auto func = [](const LambdaField_T_ field) {
    if(field.nmiss) {
      return varray_mean_mv(field.size, field.vec, field.getMissValue());
    } else {
      return varray_mean(field.size, field.vec);
    }
  };
  
  return field_operation(func, field);
}

double
field_meanw(const Field_t &field)
{
    const auto func = [](const LambdaField_T_ field) {
    if(field.nmiss) {
      return varray_weighted_mean_mv(field.size, field.vec, field.weightv, field.getMissValue());
    } else {
      return varray_weighted_mean(field.size, field.vec, field.weightv, field.getMissValue());
    }
  };
  
  return field_operation(func, field);
}

double
field_avg(const Field_t &field)
{
     const auto func = [](const LambdaField_T_ field) {
    if(field.nmiss) {
      return varray_avg_mv(field.size, field.vec, field.getMissValue());
    } else {
      return varray_mean(field.size, field.vec);
    }
  };
  
  return field_operation(func, field);
}

double
field_avgw(const Field_t &field)
{
      const auto func = [](const LambdaField_T_ field) {
    if(field.nmiss) {
      return varray_weighted_avg_mv(field.size, field.vec, field.weightv, field.getMissValue());
    } else {
      return varray_weighted_mean(field.size, field.vec, field.weightv, field.getMissValue());
    }
  };
  
  return field_operation(func, field);
}

double
field_var(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    return varray_var(field.size, field.vec,  field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_var1(const Field_t &field)
{
    const auto func = [](const LambdaField_T_ field) {
    return varray_var_1(field.size, field.vec,  field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_skew(const Field_t &field)
{
      const auto func = [](const LambdaField_T_ field) {
    return varray_skew(field.size, field.vec,  field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_kurt(const Field_t &field)
{
        const auto func = [](const LambdaField_T_ field) {
    return varray_kurt(field.size, field.vec,  field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_median(const Field_t &field)
{
        const auto func = [](const LambdaField_T_ field) {
    return varray_median(field.size, field.vec,  field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_varw(const Field_t &field)
{
          const auto func = [](const LambdaField_T_ field) {
    return varray_weighted_var(field.size, field.vec,  field.weightv, field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
field_var1w(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    return varray_weighted_var_1(field.size, field.vec,  field.weightv, field.nmiss, field.getMissValue());
  };
  
  return field_operation(func, field);
}

double
var_to_std(double rvar, double missval)
{
  if (dbl_is_equal(rvar, missval) || rvar < 0) return missval;

  return IS_NOT_EQUAL(rvar, 0) ? std::sqrt(rvar) : 0;
}

double
field_std(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    return var_to_std(field_var(field), field.missval);
  };
  
  return field_operation(func, field);
}

double
field_std1(const Field_t &field)
{
  const auto func = [](const LambdaField_T_ field) {
    return var_to_std(field_var1(field), field.missval);
  };
  
  return field_operation(func, field);
}

double
field_stdw(const Field_t &field)
{
    const auto func = [](const LambdaField_T_ field) {
    return var_to_std(field_varw(field), field.missval);
  };
  
  return field_operation(func, field);
}

double
field_std1w(const Field_t &field)
{
      const auto func = [](const LambdaField_T_ field) {
    return var_to_std(field_var1w(field), field.missval);
  };
  
  return field_operation(func, field);
}

template<typename T>
void
field_rms(const Field<T> &field, const Field<T> &field2, Field<T> &field3)
{
  size_t rnmiss = 0;
  const auto grid1 = field.grid;
  //  size_t nmiss1   = field.nmiss;
  const auto array1 = field.vec.data();
  const auto grid2 = field2.grid;
  //  size_t nmiss2   = field2.nmiss;
  const auto array2 = field2.vec.data();
  const auto missval1 = field.missval;
  const auto missval2 = field2.missval;
  const auto &w = field.weightv;
  auto rsum = 0.0, rsumw = 0.0;

  const auto len = gridInqSize(grid1);
  if (len != gridInqSize(grid2)) cdo_abort("fields have different size!");

  // if ( nmiss1 )
  {
    for (size_t i = 0; i < len; ++i)
      if (!dbl_is_equal(w[i], missval1))
        {
          rsum = ADDMN(rsum, MULMN(w[i], MULMN(SUBMN(array2[i], array1[i]), SUBMN(array2[i], array1[i]))));
          rsumw = ADDMN(rsumw, w[i]);
        }
  }
  /*
else
  {
    for ( i = 0; i < len; i++ )
      {
        rsum  += w[i] * array1[i];
        rsumw += w[i];
      }
  }
  */

  const auto ravg = SQRTMN(DIVMN(rsum, rsumw));

  if (dbl_is_equal(ravg, missval1)) rnmiss++;

  field3.vec[0] = ravg;
  field3.nmiss = rnmiss;
}

void
field_rms(const Field_t &field, const Field_t &field2, Field_t &field3) {
  field_operation3(&field_rms<float>, &field_rms<double>, field, field2, field3);
}

template <typename T>
double
array_pctl(size_t len, T *array, size_t nmiss, T missval, const double pn)
{
  double pctl = missval;

  if (len != nmiss)
    {
      if (nmiss)
        {
          Varray<T> v(len);

          size_t j = 0;
          for (size_t i = 0; i < len; ++i)
            if (!dbl_is_equal(array[i], missval)) v[j++] = array[i];

          if (nmiss != len - j)
            cdo_warning("Internal problem, inconsistent number of missing values (nmiss: exprected=%zu found=%zu!)", nmiss, len - j);

          pctl = percentile(v.data(), j, pn);
        }
      else
        {
          pctl = percentile(array, len, pn);
        }
    }

  return pctl;
}

double
field_pctl(Field_t &field, const double pn)
{
  const auto func = [](LambdaField_T_ field, const double pn) {
    return array_pctl(field.size, field.vec.data(), field.nmiss, field.getMissValue(), pn);
  };
  
  return field_operation(func, field, pn);
}

static int
compare_double(const void *const a, const void *const b)
{
  const double *const x = (const double *) a;
  const double *const y = (const double *) b;
  return ((*x > *y) - (*x < *y)) * 2 + (*x > *y) - (*x < *y);
}

template<typename T>
double
field_rank(Field<T> &field)
{
  auto res = 0.0;
  // Using first value as reference (observation)
  const auto val = field.vec[0];
  const auto array = &field.vec[1];
  const auto len = field.size - 1;

  if (field.nmiss) return field.missval;

  std::qsort(array, len, sizeof(double), compare_double);

  if (val > array[len - 1])
    res = (double) len;
  else
    for (size_t j = 0; j < len; ++j)
      if (array[j] >= val)
        {
          res = (double) j;
          break;
        }

  return res;
}

double
field_rank(Field_t &field) {
  return field_operation(&field_rank<float>, &field_rank<double>, field);
}

double
field_function(const Field_t &field, int function)
{
  // clang-format off
  switch (function)
    {
    case FieldFunc_Min:    return field_min(field);
    case FieldFunc_Max:    return field_max(field);
    case FieldFunc_Range:  return field_range(field);
    case FieldFunc_Sum:    return field_sum(field);
    case FieldFunc_Mean:   return field_mean(field);
    case FieldFunc_Avg:    return field_avg(field);
    case FieldFunc_Std:    return field_std(field);
    case FieldFunc_Std1:   return field_std1(field);
    case FieldFunc_Var:    return field_var(field);
    case FieldFunc_Var1:   return field_var1(field);
    case FieldFunc_Meanw:  return field_meanw(field);
    case FieldFunc_Avgw:   return field_avgw(field);
    case FieldFunc_Stdw:   return field_stdw(field);
    case FieldFunc_Std1w:  return field_std1w(field);
    case FieldFunc_Varw:   return field_varw(field);
    case FieldFunc_Var1w:  return field_var1w(field);
    case FieldFunc_Skew:   return field_skew(field);
    case FieldFunc_Kurt:   return field_kurt(field);
    case FieldFunc_Median: return field_median(field);
    default: cdo_abort("%s: function %d not implemented!", __func__, function);
    }
  // clang-format on
  return 0.0;
}
