/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Volker Neff

*/
#ifndef FIELD_H
#define FIELD_H

#include <cstdio>
#include "varray.h"
#include "cdo_vlist.h"

#include <variant>

/************************************************
 ************************************************
 *********** Field Declearations ****************
 ************************************************
 ***********************************************/

class FieldFactory;

template<typename T>
class Field;

using FieldFloat = Field<float>;
using FieldDouble = Field<double>;
class Field_t;

template<typename T>
class Field3D;

using Field3DFloat = Field3D<float>;
using Field3DDouble = Field3D<double>;
class Field3D_t;

using FieldVector = std::vector<Field_t>;
using FieldVector2D = std::vector<FieldVector>;
using FieldVector3D = std::vector<FieldVector2D>;

using Field3DVector = std::vector<Field3D_t>;

/**
 * @brief show that a Field<T> should be used
 * 
 * This define should show the programmer what the compiler should choos.
 * If a field_operation is used with a lambda, the parts for the Field<T> must be replaced by auto&. To show what the porgrammer had in mind this macro could be used.
 * 
 */
#define LambdaField_T_ auto&

/**
 * @brief show that a Field<T> should be used
 * 
 * This define should show the programmer what the compiler should choos.
 * If a field_operation is used with a lambda, the parts for the Field<T> must be replaced by auto&. To show what the porgrammer had in mind this macro could be used.
 * 
 */
#define LambdaField3D_T_ auto&

/************************************************
 ************************************************
 ******************** Field  ********************
 ************************************************
 ***********************************************/


template<typename T>
class // Field
#ifdef WARN_UNUSED
[[gnu::warn_unused]]
#endif
Field
{
public:
  int fpeRaised = 0;
  int nwpv = 1;         // number of words per value; real:1  complex:2
  int grid = -1;
  using memType = T;

  size_t gridsize = 0;
  size_t size = 0;
  size_t nsamp = 0;

  size_t nmiss = 0;
  double missval = 0;

  Varray<T> vec;
  Varray<double> weightv;

  // is constructed by FieldFactory.init()
  Field() = default;
  void resize(size_t count);
  void resize(size_t count, const T value);
  template<typename U>
  void resize(size_t count, const U value);
  bool empty() const;
  void check_gridsize() const;

  bool hasData() const;

  T getMissValue() const;
  double getMissValueDouble() const;
  void setMissValue(double value);

protected:
  Field(int grid, size_t gridsize, double missval, size_t size);
  size_t m_count = 0;

  // Is requiered for the FieldFactory.init function
  friend class FieldFactory;
};

/************************************************
 ************************************************
 ******************** Field3D *******************
 ************************************************
 ***********************************************/

template<typename T>
class // Field3D
#ifdef WARN_UNUSED
[[gnu::warn_unused]]
#endif
Field3D : public Field<T>
{
public:
  size_t nlevels = 0;

  Field3D() {}

private:
  Field3D(int grid, size_t gridsize, double missval, size_t size, size_t nlevels);

  // Is requiered for the FieldFactory.init function
  friend class FieldFactory;
};

/************************************************
 ************************************************
 ****************** Field_t  ********************
 ************************************************
 ***********************************************/

class Field_t : public std::variant<FieldDouble, FieldFloat> {
public:
  using _var_ = std::variant<FieldDouble, FieldFloat>;
public:
  // Overwrites of the std::variant<> constructors
  constexpr Field_t() noexcept;
  Field_t(Field_t& other);
  template< class T >
  constexpr Field_t( T&& t ) noexcept;

  template< class T,
            class... Args >
  constexpr explicit Field_t( std::in_place_type_t<T>,
                            Args&&... args );

  template< class T,
          class U,
          class... Args >
  constexpr explicit Field_t(std::in_place_type_t<T>,
                            std::initializer_list<U> il,
                            Args&&... args );

  template< std::size_t I,
            class... Args >
  constexpr explicit Field_t( std::in_place_index_t<I>,
                            Args&&... args );

  template< std::size_t I,
          class U,
          class... Args >
  constexpr explicit Field_t( std::in_place_index_t<I>,
                            std::initializer_list<U> il,
                            Args&&... args );

  // Overwrites of the std::variant<> operator=
  Field_t& operator=( const Field_t& rhs );
  Field_t& operator=( Field_t&& rhs ) noexcept;
  template< class T >
  Field_t& operator=( T&& t ) noexcept;

  // Self defined methods
  void resize(size_t count);

  template<typename U>
  void resize(size_t count, const U value);

  void setMissValue(double value);
  double getMissValue() const;

  bool empty() const;

  size_t getNumMiss() const;
  void setNumMiss(const size_t value);

  size_t getSize() const;
};

/************************************************
 ************************************************
 ****************** Field3D_t  ********************
 ************************************************
 ***********************************************/

class Field3D_t : public std::variant<Field3DDouble, Field3DFloat> {
public:
  using _var_ = std::variant<Field3DDouble, Field3DFloat>;
public:
  // Overwrites of the std::variant<> constructors
  constexpr Field3D_t() noexcept;
  Field3D_t(Field3D_t& other);
  template< class T >
  constexpr Field3D_t( T&& t ) noexcept;

  template< class T,
            class... Args >
  constexpr explicit Field3D_t( std::in_place_type_t<T>,
                            Args&&... args );

  template< class T,
          class U,
          class... Args >
  constexpr explicit Field3D_t(std::in_place_type_t<T>,
                            std::initializer_list<U> il,
                            Args&&... args );

  template< std::size_t I,
            class... Args >
  constexpr explicit Field3D_t( std::in_place_index_t<I>,
                            Args&&... args );

  template< std::size_t I,
          class U,
          class... Args >
  constexpr explicit Field3D_t( std::in_place_index_t<I>,
                            std::initializer_list<U> il,
                            Args&&... args );

  // Overwrites of the std::variant<> operator=
  Field3D_t& operator=( const Field3D_t& rhs );
  Field3D_t& operator=( Field3D_t&& rhs ) noexcept;
  template< class T >
  Field3D_t& operator=( T&& t ) noexcept;

  // Self defined methods
  void resize(size_t count);

  template<typename U>
  void resize(size_t count, const U value);

  void setMissValue(double value);
  double getMissValue() const;
};

/************************************************
 ************************************************
 ************* FieldFactory  ********************
 ************************************************
 ***********************************************/
class FieldFactory {
public:
  FieldFactory() = default;

  /**
   * @brief init a Field_t based on a CdoVar and alloc memory
   * 
   * @param var 
   * @return Field_t new Field_t that contains Field<float> or Field<double>
   */
  static Field_t init(const CdoVar &var);

  /**
   * @brief init a Field_t based on the memType
   * 
   * @param var MemType::Float or MemType::Type
   * @return Field_t new Field_t that contains Field<float> or Field<double>
   */
  static Field_t init(const MemType memType);

  /**
   * @brief init a Field_t based on a CdoVar and alloc memory. The memory type is forced by memType
   * 
   * @param var 
   * @param memType force the Type of Field_t
   * @return Field_t new Field_t that contains Field<float> or Field<double>
   */
  static Field_t init(const CdoVar &var, const MemType memType);

  /**
   * @brief init a Field3D_t based on a CdoVar and alloc memory
   * 
   * @param var 
   * @return Field3D_t new FieldField3D_t_t that contains Field3D<float> or Field3D<double>
   */
  static Field3D_t init3D(const CdoVar &var);

  /**
   * @brief init a Field3D_t based on the memType
   * 
   * @param var MemType::Float or MemType::Type
   * @return Field3D_t new Field3D_t that contains Field3D<float> or Field3D<double>
   */
  static Field3D_t init3D(const MemType memType);

  /**
   * @brief if the type of the Field_t is allready the required type only the internal vector is reallocated. If the type is different a new Field_t will constructed.
   * 
   * The function should be used like below:
   * @code {.cpp}
   * Field_t field;
   * CdoVar var;
   * field = FieldFactory::reInit(field, var);
   * @endcode
   * 
   * 
   * @param field 
   * @param var 
   * @return Field_t 
   */
  static Field_t reInit(Field_t& field, const CdoVar &var);
};

/************************************************
 ************************************************
 *********** Global Types ***********************
 ************************************************
 ***********************************************/

struct RecordInfo
{
  int varID = 0;
  int levelID = 0;
  bool lconst = false;
};

enum field_flag
{
  FIELD_VEC = 2,  // allocated memory
  FIELD_FLT = 4,  // 32-bit float
  FIELD_DBL = 8,  // 64-bit float
  FIELD_NAT = 16, // native: 32-bit float for 32-bit float data, otherweise 64-bit float
};

/************************************************
 ************************************************
 ******** Inline Template definitions ***********
 ************************************************
 ***********************************************/

#include "field.inl"
#include "field3D.inl"
#include "field_t.inl"
#include "field3D_t.inl"
#include "field_factory.inl"

#include "field_operation.h"
#include "field_operation.inl"

#endif /* FIELD_H */
