/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Uwe Schulzweida

*/

#include <algorithm>
#include <cdi.h>

#include "cdo_options.h"
#include "process_int.h"
#include "field.h"
#include "field_functions.h"

// clang-format off

// arithmetic
auto arith_add   = [](const double a, const double b) { return a + b; };
auto arith_sub   = [](const double a, const double b) { return a - b; };
auto arith_mul   = [](const double a, const double b) { return a * b; };
auto arith_div   = [](const double a, const double b) { return a / b; };
auto arith_min   = [](const auto a, const auto b) { return (b > a) ? a : b; };
auto arith_max   = [](const auto a, const auto b) { return (b < a) ? a : b; };
auto arith_sumq  = [](const double a, const double b) { return a + b * b; };
auto arith_atan2 = [](const double a, const double b) { return std::atan2(a, b); };

// arithmetic with missing values
auto arith_add_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = (isEQ(a, mv_a) || isEQ(b, mv_b)) ? mv_a : arith_add(a, b); };
auto arith_sub_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = (isEQ(a, mv_a) || isEQ(b, mv_b)) ? mv_a : arith_sub(a, b); };
auto arith_mul_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = (isEQ(a, 0) || isEQ(b, 0)) ? 0 : ((isEQ(a, mv_a) || isEQ(b, mv_b)) ? mv_a : arith_mul(a, b)); };
auto arith_div_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = (isEQ(a, mv_a) || isEQ(b, mv_b) || isEQ(b, 0)) ? mv_a : arith_div(a, b); };
auto arith_min_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = isEQ(b, mv_b) ? a : (isEQ(a, mv_a) ? b : arith_min(a, b)); };
auto arith_max_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { a = isEQ(b, mv_b) ? a : (isEQ(a, mv_a) ? b : arith_max(a, b)); };
auto arith_sum_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                    { if (!isEQ(b, mv_b)) a = (isEQ(a, mv_a) ? b : arith_add(a, b)); };
auto arith_sumq_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                     { if (!isEQ(b, mv_b)) a = (isEQ(a, mv_a) ? arith_mul(b, b) : arith_sumq(a, b)); };
auto arith_atan2_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                      { a = (isEQ(a, mv_a) || isEQ(b, mv_b)) ? mv_a : arith_atan2(a, b); };
auto arith_vinit_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                      { a = !isEQ(b, mv_b); (void)mv_a; };
auto arith_vincr_mv = [](auto &a, const auto mv_a, const auto b, const auto mv_b, auto isEQ)
                      { if (!isEQ(b, mv_b)) a = a + 1; (void)mv_a; };
// clang-format on

template <typename T1, typename T2>
static void
varray2_div(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, const double missval1)
{
  assert(len > 0);
  assert(v1.size() > 0);
  assert(v2.size() > 0);
  assert(len <= v1.size());
  assert(len <= v2.size());

  for (size_t i = 0; i < len; ++i) v1[i] = is_equal(v2[i], 0.0) ? missval1 : arith_div(v1[i], v2[i]);
}

template <typename T1, typename T2, typename FUNC>
static void
varray2_arith(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, FUNC func)
{
  assert(len > 0);
  assert(v1.size() > 0);
  assert(v2.size() > 0);
  assert(len <= v1.size());
  assert(len <= v2.size());

  for (size_t i = 0; i < len; ++i) v1[i] = func(v1[i], v2[i]);
}

template <typename T1, typename T2, typename FUNC>
static void
varray2_arith_mv(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, const double missval1, const double missval2, FUNC func)
{
  assert(len > 0);
  assert(v1.size() > 0);
  assert(v2.size() > 0);
  assert(len <= v1.size());
  assert(len <= v2.size());

  const T1 mv1 = missval1;
  const T2 mv2 = missval2;

  if (std::isnan(missval1) || std::isnan(missval2))
    {
#ifdef _OPENMP
#pragma omp parallel for default(shared) schedule(static)
#endif
      for (size_t i = 0; i < len; ++i) func(v1[i], mv1, v2[i], mv2, dbl_is_equal);
    }
  else
    {
#ifdef _OPENMP
#pragma omp parallel for default(shared) schedule(static)
#endif
      for (size_t i = 0; i < len; ++i) func(v1[i], mv1, v2[i], mv2, is_equal);
    }
}

// init valid values
void
field2_vinit(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field1.getMissValue(), arith_vinit_mv);
    field1.nmiss = field2.nmiss;
  };

  field_operation2DiffType(func, field2, field1);
}

// increment valid values
void
field2_vincr(Field_t &field1, const Field_t &field2)
{

  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_vincr_mv);
    field1.nmiss = field2.nmiss;
  };

  field_operation2DiffType(func, field2, field1);
}

// init valid values
template <typename T1, typename T2>
static void
field2_vinit(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, const T2 missval, int vinit)
{
  for (size_t i = 0; i < len; ++i) v1[i] = (DBL_IS_EQUAL(v2[i], missval)) ? 0 : vinit;
}

void
field2_vinit(Field_t &field1, const Field_t &field2, int vinit)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1, int vinit) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    using T = std::remove_reference_t<decltype(field2)>;
    field2_vinit(field1.size, field1.vec, field2.vec, static_cast<typename T::memType>(field1.getMissValue()), vinit);
    field1.nmiss = field2.nmiss;
  };

  field_operation2DiffType(func, field2, field1, vinit);
}

// increment valid values
template <typename T1, typename T2>
static void
field2_vincr(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, const T2 missval, int vincr)
{
  for (size_t i = 0; i < len; ++i)
    if (!DBL_IS_EQUAL(v2[i], missval)) v1[i] += vincr;
}

void
field2_vincr(Field_t &field1, const Field_t &field2, int vincr)
{

    const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1, int vincr) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }
    
    using T = std::remove_reference_t<decltype(field2)>;
    field2_vincr(field1.size, field1.vec, field2.vec, static_cast<typename T::memType>(field1.getMissValue()), vincr);
    field1.nmiss = field2.nmiss;
  };

  field_operation2DiffType(func, field2, field1, vincr);
}

void
field2_add(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_add_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_add);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_sum(Field_t &field1, const Field_t &field2)
{
    const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_sum_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_add);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

const auto 
field2_sumw_ = [](const LambdaField_T_ field2, LambdaField_T_ field1, double w)
{
  const auto missval1 = field1.missval;
  const auto missval2 = field2.missval;
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;

  auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field1.nmiss || field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        if (!DBL_IS_EQUAL(array2[i], missval2))
          {
            if (!DBL_IS_EQUAL(array1[i], missval1))
              array1[i] += w * array2[i];
            else
              array1[i] = w * array2[i];
          }

      field1.nmiss = field_num_miss(field1);
    }
  else
    {
#ifdef _OPENMP
#pragma omp parallel for default(shared) schedule(static)
#endif
      for (size_t i = 0; i < len; ++i) array1[i] += w * array2[i];
    }
};

void field2_sumw(Field_t &field1, const Field_t &field2, double w) {
  field_operation2DiffType(field2_sumw_, field2, field1, w);
}

/*
 * Compute the occurrence of values in field, if they do not equal refval.
 * This can be used to compute the lengths of multiple periods in a timeseries.
 * Missing field values are handled like refval, i.e. they stop a running period.
 * If there is missing data in the occurence field, missing fields values do not
 * change anything (they do not start a non-period by setting occurrence to zero).
 */
template<typename T>
void
field2_sumtr(const Field<T> &field, Field<T> &occur, const T refval)
{
  auto omissval = occur.missval;
  auto fmissval = field.missval;
  auto &oarray = occur.vec;
  const auto &farray = field.vec;

  auto len = occur.size;
  if (len != field.size) cdo_abort("Fields have different size (%s)", __func__);

  if (occur.nmiss || field.nmiss)
    {
#ifdef _OPENMP
#pragma omp parallel for default(shared) schedule(static)
#endif
      for (size_t i = 0; i < len; ++i)
        if (!DBL_IS_EQUAL(farray[i], fmissval))
          {
            if (!DBL_IS_EQUAL(oarray[i], omissval))
              oarray[i] = (DBL_IS_EQUAL(farray[i], refval)) ? 0.0 : oarray[i] + 1.0;
            else
              oarray[i] = (DBL_IS_EQUAL(farray[i], refval)) ? 0.0 : 1.0;
          }
        else
          {
            if (!DBL_IS_EQUAL(oarray[i], omissval)) oarray[i] = 0.0;
          }

      occur.nmiss = field_num_miss(occur);
    }
  else
    {
#ifdef _OPENMP
#pragma omp parallel for default(shared) schedule(static)
#endif
      for (size_t i = 0; i < len; ++i) oarray[i] = (DBL_IS_EQUAL(farray[i], refval)) ? 0.0 : oarray[i] + 1.0;
    }
}

void 
field2_sumtr(Field_t &occur, const Field_t &field, const double refval) {
  field_operation2(&field2_sumtr<float>, &field2_sumtr<double>, field, occur, refval);
}

void
field2_sumq(Field_t &field1, const Field_t &field2)
{
    const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_sumq_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_sumq);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_sumsumq(Field_t &field1, Field_t &field2, const Field_t &field3)
{
  field2_sumq(field2, field3);
  field2_sum(field1, field3);
}

template<typename T>
void
field2_sumqw(const Field<T> &field2, Field<T> &field1, const T w)
{
  const auto missval1 = field1.missval;
  const auto missval2 = field2.missval;
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field1.nmiss || field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        if (!DBL_IS_EQUAL(array2[i], missval2))
          {
            if (!DBL_IS_EQUAL(array1[i], missval1))
              array1[i] += w * array2[i] * array2[i];
            else
              array1[i] = w * array2[i] * array2[i];
          }

      field1.nmiss = field_num_miss(field1);
    }
  else
    {
      for (size_t i = 0; i < len; ++i) array1[i] += w * array2[i] * array2[i];
    }
}

void
field2_sumqw(Field_t &field1, const Field_t &field2, const double w)
{
  field_operation2(&field2_sumqw<float>, &field2_sumqw<double>, field2, field1, w);
}

void
field2_sub(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_sub_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_sub);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_mul(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_mul_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_mul);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_div(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_div_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_div);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_atan2(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_atan2_mv);
    field_num_mv(field1);
  };

  field_operation2DiffType(func, field2, field1);
}

template<typename T>
void
field2_set_miss(const Field<T> &field2, Field<T> &field1)
{
  const auto missval1 = field1.missval;
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field1.nmiss)
    {
      for (size_t i = 0; i < len; ++i) array1[i] = DBL_IS_EQUAL(array1[i], missval1) ? array2[i] : array1[i];

      field1.nmiss = field_num_miss(field1);
    }
}

void
field2_set_miss(Field_t &field1, const Field_t &field2) {
  field_operation2(&field2_set_miss<float>, &field2_set_miss<double>, field2, field1);
}

void
field2_min(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_min_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_min);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_max(Field_t &field1, const Field_t &field2)
{
 const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) {
      cdo_abort("Fields have different size (%s)", __func__);
    }

    if (field1.getMissValueDouble() || field2.getMissValueDouble()) {
      varray2_arith_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue(), arith_max_mv);
      field_num_mv(field1);
    } else {
      varray2_arith(field1.size, field1.vec, field2.vec, arith_max);
    }
  };

  field_operation2DiffType(func, field2, field1);
}

void
field2_maxmin(Field_t &field1, Field_t &field2, const Field_t &field3)
{
  field2_min(field2, field3);
  field2_max(field1, field3);
}

auto field2_set_index = [](auto &v1, auto &v2, const auto v3, const auto idx) {
  v2 = v3;
  v1 = idx;
};

template <typename T, typename U>
void
field2_minidx(Field<U> &field1, Field<U> &field2, size_t nmiss3, size_t len, double missval3, const Varray<T> &v3, int idx)
{
  const auto missval1 = field1.getMissValue();
  const auto missval2 = field2.getMissValue();
  auto &v1 = field1.vec;
  auto &v2 = field2.vec;

  if (field2.nmiss || nmiss3)
    {
      for (size_t i = 0; i < len; ++i)
        {
          if (DBL_IS_EQUAL(v3[i], missval3))
            {
              if (DBL_IS_EQUAL(v2[i], missval2)) v1[i] = missval1;
            }
          else if (v3[i] < v2[i] || DBL_IS_EQUAL(v2[i], missval2))
            {
              field2_set_index(v1[i], v2[i], v3[i], idx);
            }
        }

      field_num_mv(field1);
      field_num_mv(field2);
    }
  else
    {
      for (size_t i = 0; i < len; ++i)
        {
          if (v3[i] < v2[i]) field2_set_index(v1[i], v2[i], v3[i], idx);
        }
    }
}

template <typename T>
void
field2_minidx(size_t nmiss3, size_t len, double missval3, Field_t &field1, Field_t &field2, const Varray<T> &v3, int idx) {
  field_operation2(&field2_minidx<T, float>, &field2_minidx<T, double>, field1, field2, nmiss3, len, missval3, v3, idx);
}

void
field2_minidx(Field_t &field1, Field_t &field2, const Field_t &field3, int idx)
{
  const auto func = [](const LambdaField_T_ field3,  LambdaField_T_ field1, LambdaField_T_ field2, int idx) {
    if (field1.size != field3.size) cdo_abort("Fields have different size (%s)", __func__);
    if (field2.size != field3.size) cdo_abort("Fields have different size (%s)", __func__);

    field2_minidx(field1, field2, field3.nmiss, field3.size, field3.missval, field3.vec, idx);
  };

  field_operation3(func, field3, field1, field2, idx);
}

template <typename T, typename U>
void
field2_maxidx(Field<U> &field1, Field<U> &field2, size_t nmiss3, size_t len, double missval3, const Varray<T> &v3, int idx)
{
  const auto missval1 = field1.missval;
  const auto missval2 = field2.missval;
  auto &v1 = field1.vec;
  auto &v2 = field2.vec;

  if (field2.nmiss || nmiss3)
    {
      for (size_t i = 0; i < len; ++i)
        {
          if (DBL_IS_EQUAL(v3[i], missval3))
            {
              if (DBL_IS_EQUAL(v2[i], missval2)) v1[i] = missval1;
            }
          else if (v3[i] > v2[i] || DBL_IS_EQUAL(v2[i], missval2))
            {
              field2_set_index(v1[i], v2[i], v3[i], idx);
            }
        }

      field_num_mv(field1);
      field_num_mv(field2);
    }
  else
    {
      for (size_t i = 0; i < len; ++i)
        {
          if (v3[i] > v2[i]) field2_set_index(v1[i], v2[i], v3[i], idx);
        }
    }
}

template <typename T>
void
field2_maxidx(size_t nmiss3, size_t len, double missval3, Field_t &field1, Field_t &field2, const Varray<T> &v3, int idx) {
  field_operation2(&field2_maxidx<T, float>, &field2_maxidx<T, double>, field1, field2, nmiss3, len, missval3, v3, idx);
}

void
field2_maxidx(Field_t &field1, Field_t &field2, const Field_t &field3, int idx)
{
   const auto func = [](const LambdaField_T_ field3,  LambdaField_T_ field1, LambdaField_T_ field2, int idx) {
    if (field1.size != field3.size) cdo_abort("Fields have different size (%s)", __func__);
    if (field2.size != field3.size) cdo_abort("Fields have different size (%s)", __func__);

    field2_maxidx(field1, field2, field3.nmiss, field3.size, field3.missval, field3.vec, idx);
  };

  field_operation3(func, field3, field1, field2, idx);
}

template<typename T>
static size_t
field_set_nmiss(const size_t len, Varray<T> &v, T missval)
{
  size_t nmiss = 0;

  if (std::isnan(missval))
    {
      for (size_t i = 0; i < len; ++i)
        if (DBL_IS_EQUAL(v[i], missval) || v[i] < 0)
          {
            v[i] = missval;
            nmiss++;
          }
    }
  else
    {
      for (size_t i = 0; i < len; ++i)
        if (IS_EQUAL(v[i], missval) || v[i] < 0)
          {
            v[i] = missval;
            nmiss++;
          }
    }

  return nmiss;
}

template<typename T>
void
field2_var(const Field<T> &field2, const Field<T> &field3, Field<T> &field1, const int divisor)
{
  const auto missval1 = field1.getMissValue();
  const auto missval2 = field2.getMissValue();
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;
  const auto &array3 = field3.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field1.nmiss || field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        {
          const auto temp = DIVMN(MULMN(array1[i], array1[i]), array3[i]);
          array1[i] = DIVMN(SUBMN(array2[i], temp), array3[i] - divisor);
          if (array1[i] < 0 && array1[i] > -1.e-5) array1[i] = 0;
        }
    }
  else
    {
      for (size_t i = 0; i < len; ++i)
        {
          const auto temp = DIV(MUL(array1[i], array1[i]), array3[i]);
          array1[i] = DIV(SUB(array2[i], temp), array3[i] - divisor);
          if (array1[i] < 0 && array1[i] > -1.e-5) array1[i] = 0;
        }
    }

  field1.nmiss = field_set_nmiss(len, array1, missval1);
}

void
field2_var(Field_t &field1, const Field_t &field2, const Field_t &field3, const int divisor) {
  field_operation3(&field2_var<float>, &field2_var<double>, field2, field3, field1, divisor);
}

template<typename T>
void
field2_std(const Field<T> &field2, const Field<T> &field3, Field<T> &field1, const int divisor)
{
  const auto missval1 = field1.missval;
  auto &array1 = field1.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  field2_var(field2, field3, field1, divisor);

  size_t nmiss = 0;
  for (size_t i = 0; i < len; ++i)
    if (DBL_IS_EQUAL(array1[i], missval1) || array1[i] < 0)
      {
        array1[i] = missval1;
        nmiss++;
      }
    else
      {
        array1[i] = IS_NOT_EQUAL(array1[i], 0) ? std::sqrt(array1[i]) : 0;
      }
  field1.nmiss = nmiss;
}

void
field2_std(Field_t &field1, const Field_t &field2, const Field_t &field3, const int divisor) {
  field_operation3(&field2_std<float>, &field2_std<double>, field2, field3, field1, divisor);
}

template<typename T>
void
fieldc_var(const Field<T> &field2, Field<T> &field1, const int nsets, const int divisor)
{
  const auto nsetx = nsets - divisor;
  const auto missval1 = field1.getMissValue();
  const auto missval2 = field2.getMissValue();
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;
  double temp;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (nsetx == 0)
    {
      for (size_t i = 0; i < len; ++i) array1[i] = missval1;
    }
  else if (field1.nmiss || field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        {
          temp = MULMN(array1[i], array1[i]) / nsets;
          array1[i] = SUBMN(array2[i], temp) / nsetx;
          if (array1[i] < 0 && array1[i] > -1.e-5) array1[i] = 0;
        }
    }
  else
    {
      for (size_t i = 0; i < len; ++i)
        {
          temp = MUL(array1[i], array1[i]) / nsets;
          array1[i] = SUB(array2[i], temp) / nsetx;
          if (array1[i] < 0 && array1[i] > -1.e-5) array1[i] = 0;
        }
    }

  field1.nmiss = field_set_nmiss(len, array1, missval1);
}

void
fieldc_var(Field_t &field1, const Field_t &field2, const int nsets, const int divisor) {
  field_operation2(&fieldc_var<float>, &fieldc_var<double>, field2, field1, nsets, divisor);
}

template<typename T>
void
fieldc_std(const Field<T> &field2, Field<T> &field1, const int nsets, const int divisor)
{
  const auto missval1 = field1.missval;
  auto &array1 = field1.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  fieldc_var(field2, field1, nsets, divisor);

  size_t nmiss = 0;
  for (size_t i = 0; i < len; ++i)
    if (DBL_IS_EQUAL(array1[i], missval1) || array1[i] < 0)
      {
        array1[i] = missval1;
        nmiss++;
      }
    else
      {
        array1[i] = IS_NOT_EQUAL(array1[i], 0) ? std::sqrt(array1[i]) : 0;
      }

  field1.nmiss = nmiss;
}

void
fieldc_std(Field_t &field1, const Field_t &field2, const int nsets, const int divisor) {
  field_operation2(&fieldc_std<float>, &fieldc_std<double>, field2, field1, nsets, divisor);
}

template<typename T>
void
field2_moq(const Field<T> &field2, Field<T> &field1)
{
  const auto missval1 = field1.missval;
  const auto missval2 = field2.missval;
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        if (!DBL_IS_EQUAL(array2[i], missval2))
          array1[i] = array2[i] * array2[i];
        else
          array1[i] = missval1;

      field1.nmiss = field_num_miss(field1);
    }
  else
    {
      for (size_t i = 0; i < len; ++i) array1[i] = array2[i] * array2[i];
    }
}

void
field2_moq(Field_t &field1, const Field_t &field2) {
  field_operation2(&field2_moq<float>, &field2_moq<double>, field2, field1);
}

template<typename T>
void
field2_moqw(const Field<T> &field2, Field<T> &field1, const double w)
{
  const auto missval1 = field1.missval;
  const auto missval2 = field2.missval;
  auto &array1 = field1.vec;
  const auto &array2 = field2.vec;

  const auto len = field1.size;
  if (len != field2.size) cdo_abort("Fields have different size (%s)", __func__);

  if (field2.nmiss)
    {
      for (size_t i = 0; i < len; ++i)
        if (!DBL_IS_EQUAL(array2[i], missval2))
          array1[i] = w * array2[i] * array2[i];
        else
          array1[i] = missval1;

      field1.nmiss = field_num_miss(field1);
    }
  else
    {
      for (size_t i = 0; i < len; ++i) array1[i] = w * array2[i] * array2[i];
    }
}

void
field2_moq(Field_t &field1, const Field_t &field2, const double w) {
  field_operation2(&field2_moqw<float>, &field2_moqw<double>, field2, field1, w);
}

/**
 * Counts the number of nonmissing values. The result of the operation is computed according to the following rules:
 *
 * field1  field2  result
 * a       b       a + 1
 * a       miss    a
 * miss    b       1
 * miss    miss    miss
 *
 * @param field1 the 1st input field, also holds the result
 * @param field2 the 2nd input field
 */
template <typename T1, typename T2>
static void
varray2_count_mv(const size_t len, Varray<T1> &v1, const Varray<T2> &v2, const double mv1, const double mv2)
{
  assert(len > 0);
  assert(v1.size() > 0);
  assert(v2.size() > 0);
  assert(len <= v1.size());
  assert(len <= v2.size());

  for (size_t i = 0; i < len; ++i)
    if (!DBL_IS_EQUAL(v2[i], mv2))
      {
        if (!DBL_IS_EQUAL(v1[i], mv1))
          v1[i] += 1.0;
        else
          v1[i] = 1.0;
      }
}

template <typename T>
static void
varray2_count(const size_t len, Varray<T> &v)
{
  assert(len > 0);

  for (size_t i = 0; i < len; ++i)  v[i] += 1.0;
}

void
field2_count(Field_t &field1, const Field_t &field2)
{
  const auto func = [](const LambdaField_T_ field2, LambdaField_T_ field1) {
    if (field1.size != field2.size) cdo_abort("Fields have different size (%s)", __func__);

    if(field1.nmiss || field2.nmiss) {
      varray2_count_mv(field1.size, field1.vec, field2.vec, field1.getMissValue(), field2.getMissValue());
    } else {
      varray2_count(field1.size, field1.vec);
    }
  };

  field_operation2(func, field2, field1);
}

void
field2_function(Field_t &field1, const Field_t &field2, int function)
{
  // clang-format off
  switch (function)
    {
    case FieldFunc_Add:     field2_add(field1, field2);   break;
    case FieldFunc_Min:     field2_min(field1, field2);   break;
    case FieldFunc_Max:     field2_max(field1, field2);   break;
    case FieldFunc_Sum:     field2_sum(field1, field2);   break;
    case FieldFunc_Mean:    field2_sum(field1, field2);   break;
    case FieldFunc_Avg:     field2_add(field1, field2);   break;
    case FieldFunc_Sub:     field2_sub(field1, field2);   break;
    case FieldFunc_Mul:     field2_mul(field1, field2);   break;
    case FieldFunc_Div:     field2_div(field1, field2);   break;
    case FieldFunc_Atan2:   field2_atan2(field1, field2); break;
    case FieldFunc_Setmiss: field2_set_miss(field1, field2); break;
    default: cdo_abort("%s: function %d not implemented!", __func__, function);
    }
  // clang-format on
}
