/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author:Volker Neff

*/
#ifndef FIELD_OPERATION_H
#define FIELD_OPERATION_H

#include "field.h"

/************************************************
 ************************************************
 *********** Global basis functions *************
 ************************************************
 ***********************************************/

/**
 * @brief Test that both fileds hold the same type.
 * 
 * A Field_t can hold a FieldFloat or a FieldDouble. The type is decide at runtime. This function checks that both fields hold the same type.
 * 
 * @param field1 
 * @param field2 
 * @return true if field1 and field2 hold the same type
 * @return false if field1 and field2 hold different types
 */
inline bool field_have_same_type(const Field_t& field1, const Field_t& field2);
inline bool field_have_same_type(const Field3D_t& field1, const Field3D_t& field2);
inline bool field_have_same_type(const Field3D_t& field1, const Field_t& field2);
inline bool field_have_same_type(const Field_t& field1, const Field3D_t& field2);

template< typename T >
inline constexpr bool check_T_is_field_t();
template< typename T >
inline constexpr bool check_T_is_field3D_t();
template< typename T >
inline constexpr bool check_T_is_fielX_t();

/** 
 * @brief A helper Type to detect Field<T> or Field3D<T> at compile time
 * 
 * You can find the specified structs in the field_operation.inl
 */
template<typename T>
struct FieldX_t {
  using typeFloat = void;
  using typeDouble = void;
};

template<>
struct FieldX_t<Field_t> {
  using typeFloat = FieldFloat;
  using typeDouble = FieldDouble;
};

template<>
struct FieldX_t<const Field_t> {
  using typeFloat = FieldFloat;
  using typeDouble = FieldDouble;
};

template<>
struct FieldX_t<Field3D_t> {
  using typeFloat = Field3DFloat;
  using typeDouble = Field3DDouble;
};

template<>
struct FieldX_t<const Field3D_t> {
  using typeFloat = Field3DFloat;
  using typeDouble = Field3DDouble;
};


// clang-format off
/**
 * @brief Execute a function on the field. 
 * @anchor field_operation_twoFunction
 * 
 * This function is a wrapper to prevent boiler plate code in the project. The two arguments ff and fd need the same signature excluded the first parameter. A typical signature could look like:
 * @code {.cpp}
 * int foo_f(Field<float>& field, float c);
 * int foo_d(Field<double>& field, float c);
 * 
 * Field_t field; // Could store FieldFloat or FieldDouble
 * int ret = field_operation(&foo_f, &foo_d, field, 0.42f);
 * @endcode
 * 
 * @tparam FunctionFloat Needs the same signature as FunctionDouble exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam FunctionDouble Needs the same signature as FunctionFloat exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam Field_T_ must be from type 'Field_t' or 'const Field_t'
 * @tparam ARGS... a variable number of argumentes for ff and fd
 * @param ff a function that is executed with field as first parameter if field hold a FieldFloat type
 * @param fd a function that is executed with field as first parameter if field hold a FieldDouble type
 * @param field the first parameter that is passed by reference to ff or fd depend on the type that is hold.
 * @param args additional arguments for ff or fd.
 * @return decltype(ff(std::declval<FieldFloat&>(), args...)) the return type of ff and fd. 
 */
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T_,
  typename... ARGS
  >
inline auto
field_operation(
  FunctionFloat ff, FunctionDouble fd,
  Field_T_& field,
  ARGS... args
  ) 
-> decltype(ff(std::declval<typename FieldX_t<Field_T_>::typeFloat&>(), args...));
//clang-format on

// clang-format off
/**
 * @brief Execute a function on the field. 
 * @anchor field_operation_twoFunction
 * 
 * This function is a wrapper to prevent boiler plate code in the project. The two arguments ff and fd need the same signature excluded the first parameter. A typical signature could look like:
 * @code {.cpp}
 * int foo_f(Field<float>& field1, Field<float>& field2, float c);
 * int foo_d(Field<double>& field1, Field<double>& field2, float c);
 * 
 * Field_t field; // Could store FieldFloat or FieldDouble
 * int ret = field_operation(&foo_f, &foo_d, field1, field2, 0.42f);
 * @endcode
 * 
 * @tparam FunctionFloat Needs the same signature as FunctionDouble exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam FunctionDouble Needs the same signature as FunctionFloat exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam Field_T1_ must be from type 'Field_t' or 'const Field_t'
 * @tparam Field_T2_ must be from type 'Field_t' or 'const Field_t'
 * @tparam ARGS... a variable number of argumentes for ff and fd
 * @param ff a function that is executed with field as first parameter if field hold a FieldFloat type
 * @param fd a function that is executed with field as first parameter if field hold a FieldDouble type
 * @param field the first parameter that is passed by reference to ff or fd depend on the type that is hold.
 * @param args additional arguments for ff or fd.
 * @return decltype(ff(std::declval<FieldFloat&>(), args...)) the return type of ff and fd. 
 */
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2(FunctionFloat ff, FunctionDouble fd,
Field_T1_& field1, Field_T2_& field2,
ARGS... args) 
-> decltype(ff(
                std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(),
                args...));
//clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2DiffType(
  Function f,
  const Field_T1_& field1, const Field_T2_& field2,
  ARGS... args
  ) 
-> decltype(f(
              std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(),
              args...));
//clang-format on

// clang-format off
/**
 * @brief Execute a function on the field. 
 * @anchor field_operation_twoFunction
 * 
 * This function is a wrapper to prevent boiler plate code in the project. The two arguments ff and fd need the same signature excluded the first parameter. A typical signature could look like:
 * @code {.cpp}
 * int foo_f(Field<float>& field1, Field<float>& field2, Field<float>& field3, float c);
 * int foo_d(Field<double>& field1, Field<double>& field2, Field<double>& field3, float c);
 * 
 * Field_t field; // Could store FieldFloat or FieldDouble
 * int ret = field_operation(&foo_f, &foo_d, field1, field2, field3, 0.42f);
 * @endcode
 * 
 * @tparam FunctionFloat Needs the same signature as FunctionDouble exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam FunctionDouble Needs the same signature as FunctionFloat exclude the Field<T> type. For an example look in the full function documentaion.
 * @tparam Field_T1_ must be from type 'Field_t' or 'const Field_t'
 * @tparam Field_T2_ must be from type 'Field_t' or 'const Field_t'
 * @tparam Field_T3_ must be from type 'Field_t' or 'const Field_t'
 * @tparam ARGS... a variable number of argumentes for ff and fd
 * @param ff a function that is executed with field as first parameter if field hold a FieldFloat type
 * @param fd a function that is executed with field as first parameter if field hold a FieldDouble type
 * @param field the first parameter that is passed by reference to ff or fd depend on the type that is hold.
 * @param args additional arguments for ff or fd.
 * @return decltype(ff(std::declval<FieldFloat&>(), args...)) the return type of ff and fd. 
 */
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T1_,
  typename Field_T2_,
  typename Field_T3_,
  typename... ARGS
  >
inline auto
field_operation3(
  FunctionFloat ff, FunctionDouble fd,
  Field_T1_& field1, Field_T2_& field2, Field_T3_& field3,
  ARGS... args
  ) 
-> decltype(ff(
                std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(),
                std::declval<typename FieldX_t<Field_T3_>::typeFloat&>(),
                args...));
//clang-format on

// clang-format off
/**
 * @brief Execute a function on the field.
 * @anchor field_operation_oneFunction
 * 
 * This function is a wrapper to prevent boiler plate code in the project. You can pass a function that could evaluated with FieldFloat or FieldDouble. You can also pass aditional paramter for that function. The typical way is to use a template function like below: 
 * @code {.cpp}
 * auto foo = [](auto& field, const float c) -> return int {
 *  return 42;
 * }
 * 
 * Field_t field; // Could store FieldFloat or FieldDouble
 * int ret = field_operation(foo, field, 0.42f);
 * @endcode
 * 
 * If you can't use a template function you can also use @ref field_operation_twoFunction "field_operation". This function take two functions, for the case that field hold a FieldFloat or an alternative for the case field holds a FieldDouble
 * 
 * 
 * @tparam Function Type of the function that should be executed with field. For an example look in the full function documentaion.
 * @tparam ARGS additional function arguments types for f
 * @param f the function that is executed with field
 * @param field the field that is evaluated by f
 * @param args additional function arguments for f
 * @return  -> decltype(f(std::declval<FieldFloat&>(), args...)) The return type of f.
 */
template<
  typename Function,
  typename Field_T_,
  typename... ARGS
  >
inline auto
field_operation(Function f, Field_T_& field, ARGS... args) 
-> decltype(f(std::declval<typename FieldX_t<Field_T_>::typeFloat&>(), args...));
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2(Function f, Field_T1_& field1, Field_T2_& field2, ARGS... args) 
-> decltype(f(
                std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
                args...));
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename Field_T3_,
  typename... ARGS
  >
inline auto
field_operation3(Function f, Field_T1_& field1, Field_T2_& field2, Field_T3_& field3, ARGS... args) 
-> decltype(f(
              std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T3_>::typeFloat&>(), 
              args...));
// clang-format on


/************************************************
 ************************************************
 *********** Global other functions *************
 ************************************************
 ***********************************************/



double var_to_std(double rvar, double missval);

inline void field_fill(Field_t &field, double value) {
  auto func = [](LambdaField_T_ field, double value) {
    std::fill(field.vec.begin(), field.vec.end(), value);
  };

  field_operation(func, field, value);
}

void field_copy(const Field_t &field_src, Field_t &field_tgt);
void field_copy(const Field3D_t &field_src, Field3D_t &field_tgt);
void field_copy(const Field3D_t &field_src, int levelID, Field_t &field_tgt);
void field_add(Field_t &field1, const Field3D_t &field2, int levelID);
size_t field_num_miss(const Field_t &field);
size_t field_num_mv(Field_t &field);
template<typename T>
size_t field_num_mv(Field<T> &field);
MinMax field_min_max(const Field_t &field);

template<typename T>
inline size_t field_num_mv(Field<T> &field) {
  field.nmiss = varray_num_mv(field.size, field.vec, field.getMissValue());
  return field.nmiss;
}



template <class UnaryOperation>
inline void field_transform(const Field_t &fieldIn, Field_t &fieldOut, UnaryOperation unary_op)
{
  if(!field_have_same_type(fieldIn, fieldOut)) {
    throw std::runtime_error("FieldIn and FieldOut are not from the same type");
  }

  auto func = [unary_op](const auto& fieldIn, auto& fieldOut) {
    varray_transform(fieldIn.vec, fieldOut.vec, unary_op);
  };

  field_operation(func, fieldIn, fieldOut);
}

template<typename T>
inline bool field_hold_complex_values(const Field<T>& field) {
  return field.nwpv == 2;
}

inline bool field_hold_complex_values(const Field_t& field) {
  return field_operation(&field_hold_complex_values<float>, &field_hold_complex_values<double>, field);
}

#endif // FIELD_OPERATION_H