/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Volker Neff

*/
#ifndef FIELD_OPERATION_INL
#define FIELD_OPERATION_INL

#include "field_operation.h"

/************************************************
 ************************************************
 *********** Global basis functions *************
 ************************************************
 ***********************************************/

inline bool field_have_same_type(const Field_t& field1, const Field_t& field2) {
  if(std::holds_alternative<FieldFloat>(field1) &&
     std::holds_alternative<FieldFloat>(field2)) {
    return true;
  } else if(std::holds_alternative<FieldDouble>(field1) &&
            std::holds_alternative<FieldDouble>(field2)) {
    return true;
  } else {
    return false;
  }
}

inline bool field_have_same_type(const Field3D_t& field1, const Field3D_t& field2) {
  if(std::holds_alternative<Field3DFloat>(field1) &&
     std::holds_alternative<Field3DFloat>(field2)) {
    return true;
  } else if(std::holds_alternative<Field3DDouble>(field1) &&
            std::holds_alternative<Field3DDouble>(field2)) {
    return true;
  } else {
    return false;
  }
}

inline bool field_have_same_type(const Field3D_t& field1, const Field_t& field2) {
  return false;
}

inline bool field_have_same_type(const Field_t& field1, const Field3D_t& field2) {
  return false;
}

template< typename T >
inline constexpr bool check_T_is_field_t() {
  constexpr bool is_field_t = std::is_same<T, Field_t>::value;
  constexpr bool is_const_field_t = std::is_same<T, const Field_t>::value;

  return is_field_t || is_const_field_t;
}

template< typename T >
inline constexpr bool check_T_is_field3D_t() {
  constexpr bool is_field3D_t = std::is_same<T, Field3D_t>::value;
  constexpr bool is_const_field3D_t = std::is_same<T, const Field3D_t>::value;

  return is_field3D_t || is_const_field3D_t;
}

template< typename T >
inline constexpr bool check_T_is_fielX_t() {
  constexpr bool is_field_t = check_T_is_field_t<T>();
  constexpr bool is_field3D_t = check_T_is_field3D_t<T>();

  return is_field_t || is_field3D_t;
}

// clang-format off
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T_,
  typename... ARGS
  >
inline auto
field_operation(
  FunctionFloat functionFloat, FunctionDouble functionDouble,
  Field_T_& field,
  ARGS... args
  ) 
-> decltype(functionFloat(std::declval<typename FieldX_t<Field_T_>::typeFloat&>(), args...)) {

  using typeFloat  = typename FieldX_t<Field_T_>::typeFloat;
  using typeDouble = typename FieldX_t<Field_T_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloat = decltype(functionFloat(std::declval<typeFloat&>(), args...));
  using returnTypeFunctionDouble = decltype(functionDouble(std::declval<typeDouble&>(), args...));
  static_assert(std::is_same<returnTypeFunctionFloat, returnTypeFunctionDouble>::value,
  "function float and functionDouble has not the same return types");

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T_>(), "field must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'.");

  if(std::holds_alternative<typeFloat>(field)) {
    return functionFloat(std::get<typeFloat>(field), args...);
  } else if(std::holds_alternative<typeDouble>(field)) {
    return functionDouble(std::get<typeDouble>(field), args...);
  } else {
    throw std::runtime_error("Not supported Type");
  }
}
//clang-format on


// clang-format off
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2(FunctionFloat functionFloat, FunctionDouble functionDouble,
Field_T1_& field1, Field_T2_& field2,
ARGS... args) 
-> decltype(functionFloat(
                std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
                args...)) {

  using typeFloat  = typename FieldX_t<Field_T1_>::typeFloat;
  using typeDouble = typename FieldX_t<Field_T1_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloat = decltype(functionFloat(std::declval<typeFloat&>(), std::declval<typeFloat&>(), args...));
  using returnTypeFunctionDouble = decltype(functionDouble(std::declval<typeDouble&>(), std::declval<typeDouble&>(), args...));
  static_assert(std::is_same<returnTypeFunctionFloat, returnTypeFunctionDouble>::value,
  "function float and functionDouble has not the same return types");

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T1_>(), "field1 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T2_>(), "field2 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");

  // check that both fields holds the same type
  if(!field_have_same_type(field1, field2)) {
    throw std::runtime_error("field1 and field2 does not hold the same type");
  }

  if(std::holds_alternative<typeFloat>(field1)) {
    return functionFloat(std::get<typeFloat>(field1), std::get<typeFloat>(field2), args...);
  } else if(std::holds_alternative<typeDouble>(field1)) {
    return functionDouble(std::get<typeDouble>(field1), std::get<typeDouble>(field2), args...);
  } else {
    throw std::runtime_error("Not supported Type");
  }
}
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2DiffType(
  Function f,
  Field_T1_& field1, Field_T2_& field2,
  ARGS... args
  ) 
-> decltype(f(
              std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
              args...)) {

  using typeFloat  = typename FieldX_t<Field_T1_>::typeFloat;
  using typeDouble = typename FieldX_t<Field_T1_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloatFloat = decltype(f(std::declval<typeFloat&>(), std::declval<typeFloat&>(), args...));
  using returnTypeFunctionFloatDouble = decltype(f(std::declval<typeFloat&>(), std::declval<typeDouble&>(), args...));
  using returnTypeFunctionDoubleDouble = decltype(f(std::declval<typeDouble&>(), std::declval<typeDouble&>(), args...));
  using returnTypeFunctionDoubleFloat = decltype(f(std::declval<typeDouble&>(), std::declval<typeFloat&>(), args...));
  static_assert(
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionFloatDouble>::value &&
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionDoubleDouble>::value &&
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionDoubleFloat>::value,
  "f needs the same return types for all input types");

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T1_>(), "field1 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T2_>(), "field2 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");

  if(std::holds_alternative<typeFloat>(field1)) {
    if(std::holds_alternative<typeFloat>(field2)) {
      return f(std::get<typeFloat>(field1), std::get<typeFloat>(field2),  args...);
    } else if (std::holds_alternative<typeDouble>(field2)) {
      return f(std::get<typeFloat>(field1), std::get<typeDouble>(field2),  args...);
    } else {
      throw std::runtime_error("Not supported Type of field2");
    }
  } else if(std::holds_alternative<typeDouble>(field1)) {
    if(std::holds_alternative<typeFloat>(field2)) {
      return f(std::get<typeDouble>(field1), std::get<typeFloat>(field2),  args...);
    } else if (std::holds_alternative<typeDouble>(field2)) {
      return f(std::get<typeDouble>(field1), std::get<typeDouble>(field2),  args...);
    } else {
      throw std::runtime_error("Not supported Type of field2");
    }
  } else {
    throw std::runtime_error("Not supported Type of field1");
  }
}
// clang-format on

// clang-format off
template<
  typename FunctionFloat,
  typename FunctionDouble,
  typename Field_T1_,
  typename Field_T2_,
  typename Field_T3_,
  typename... ARGS
  >
inline auto
field_operation3(
  FunctionFloat functionFloat, FunctionDouble functionDouble,
  Field_T1_& field1, Field_T2_& field2, Field_T3_& field3,
  ARGS... args
  ) 
-> decltype(functionFloat(
                std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
                std::declval<typename FieldX_t<Field_T3_>::typeFloat&>(),
                args...)) {

  using typeFloat  = typename FieldX_t<Field_T1_>::typeFloat;
  using typeDouble = typename FieldX_t<Field_T1_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloat = decltype(functionFloat(std::declval<typeFloat&>(), std::declval<typeFloat&>(), std::declval<typeFloat&>(), args...));
  using returnTypeFunctionDouble = decltype(functionDouble(std::declval<typeDouble&>(), std::declval<typeDouble&>(), std::declval<typeDouble&>(), args...));
  static_assert(std::is_same<returnTypeFunctionFloat, returnTypeFunctionDouble>::value,
  "function float and functionDouble has not the same return types");

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T1_>(), "field1 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T2_>(), "field2 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T3_>(), "field3 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");

  if(!field_have_same_type(field1, field2) ||
     !field_have_same_type(field2, field3)) {
    throw std::runtime_error("field1 and field2 does not hold the same type");
  }

  if(std::holds_alternative<typeFloat>(field1)) {
    return functionFloat(std::get<typeFloat>(field1), std::get<typeFloat>(field2), std::get<typeFloat>(field3), args...);
  } else if(std::holds_alternative<typeDouble>(field1)) {
    return functionDouble(std::get<typeDouble>(field1), std::get<typeDouble>(field2), std::get<typeDouble>(field3),args...);
  } else {
    throw std::runtime_error("Not supported Type");
  }
}
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T_,
  typename... ARGS
  >
inline auto
field_operation(Function f, Field_T_& field, ARGS... args) 
-> decltype(f(std::declval<typename FieldX_t<Field_T_>::typeFloat&>(), args...)){
  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T_>(), "field must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");

  return field_operation(f, f, field, args...);
}
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename... ARGS
  >
inline auto
field_operation2(Function f, Field_T1_& field1, Field_T2_& field2, ARGS... args) 
-> decltype(f(
              std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
              args...)){

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T1_>(), "field1 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T2_>(), "field2 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  
  return field_operation2(f, f, field1, field2, args...);
}
// clang-format on

// clang-format off
template<
  typename Function,
  typename Field_T1_,
  typename Field_T2_,
  typename Field_T3_,
  typename... ARGS
  >
inline auto
field_operation3(Function f, Field_T1_& field1, Field_T2_& field2, Field_T3_& field3, ARGS... args) 
-> decltype(f(
              std::declval<typename FieldX_t<Field_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T2_>::typeFloat&>(), 
              std::declval<typename FieldX_t<Field_T3_>::typeFloat&>(), 
              args...)){

  // Fail fast if field is not from type `Field_t` or `const Field_t`
  static_assert(check_T_is_fielX_t<Field_T1_>(), "field1 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T2_>(), "field2 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_fielX_t<Field_T3_>(), "field3 must be from type 'Field_t' or 'const Field_t' or 'Field3D_t' or 'const Field3D_t'");

  return field_operation3(f, f, field1, field2, field2, args...);
}
// clang-format on

#endif // FIELD_OPERATION_INL