/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Volker Neff

*/
#ifndef FIELD_t_INL
#define FIELD_t_INL

#include "field.h"
#include "field_operation.h"

inline constexpr Field_t::Field_t() noexcept
    : _var_()
{
}

inline Field_t::Field_t(Field_t& other)
    : _var_(other)
{
}

template< class T >
inline constexpr Field_t::Field_t( T&& t ) noexcept
    : _var_(std::move(t))
{
}

template<   class T,
            class... Args >
inline constexpr Field_t::Field_t( std::in_place_type_t<T>,
                        Args&&... args )
    : _var_(std::in_place_type_t<T>{}, args...)
{
}

template< class T,
        class U,
        class... Args >
inline constexpr Field_t::Field_t(std::in_place_type_t<T>,
                        std::initializer_list<U> il,
                        Args&&... args )
    : _var_(std::in_place_type_t<T>{}, il, args...)
{
}

template< std::size_t I,
        class... Args >
inline constexpr Field_t::Field_t( std::in_place_index_t<I>,
                        Args&&... args )
    : _var_(std::in_place_index_t<I>{}, args...)
{
}

template< std::size_t I,
        class U,
        class... Args >
inline constexpr Field_t::Field_t( std::in_place_index_t<I>,
                        std::initializer_list<U> il,
                        Args&&... args )
    : _var_(std::in_place_index_t<I>{}, il, args...)
{
}

// Overwrites of the std::variant<> operator=
inline Field_t& Field_t::operator=( const Field_t& rhs ) {
    _var_::operator=(rhs);
    return *this;
}

inline Field_t& Field_t::operator=( Field_t&& rhs ) noexcept {
    _var_::operator=(std::move(rhs));
    return *this;
}

template< class T >
inline Field_t& Field_t::operator=( T&& t ) noexcept {
    _var_::operator=(std::move(t));
    return *this;
}

inline void Field_t::resize(size_t count) {
    field_operation([](LambdaField_T_ field, size_t count){field.resize(count);}, *this, count);
}

template<typename U>
inline void Field_t::resize(size_t count, const U value) {
    field_operation([](LambdaField_T_ field, size_t count, const U value){field.resize(count, value);}, *this, count, value);
}

inline void Field_t::setMissValue(double value) {
    if(std::holds_alternative<FieldFloat>(*this)) {
        std::get<FieldFloat>(*this).setMissValue(value);
    } else if(std::holds_alternative<FieldDouble>(*this)) {
        std::get<FieldDouble>(*this).setMissValue(value);
    } else {
        throw std::runtime_error("Not supported Type");
    }
}

inline double Field_t::getMissValue() const {
    return field_operation([](LambdaField_T_ field){return field.getMissValueDouble();}, *this);
}

inline bool Field_t::empty() const {
    return field_operation([](LambdaField_T_ field){return field.empty();}, *this);
}

inline size_t Field_t::getNumMiss() const {
    return field_operation([](LambdaField_T_ field){return field.nmiss;}, *this);
}

inline void Field_t::setNumMiss(const size_t value) {
    return field_operation([](LambdaField_T_ field, const size_t value) {field.nmiss = value;}, *this, value);
}

inline size_t Field_t::getSize() const {
    return field_operation([](LambdaField_T_ field){return field.size;}, *this);
}

#endif // FIELD_t_INL